@@ -4,10 +4,10 @@
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
  * Copyright (c) 2009      Oak Ridge National Labs.  All rights reserved.
- * Copyright (c) 2010      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2010-2017 Cisco Systems, Inc.  All rights reserved
  * Copyright (c) 2013      Los Alamos National Security, LLC.  All rights
  *                         reserved.
- * Copyright (c) 2015      Research Organization for Information Science
+ * Copyright (c) 2015-2018 Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
  *
@@ -36,6 +36,7 @@
 
 #include "ompi/constants.h"
 #include "opal/datatype/opal_convertor.h"
+#include "opal/util/output.h"
 #include "mpi.h"
 
 BEGIN_C_DECLS
@@ -43,6 +44,8 @@ BEGIN_C_DECLS
 /* These flags are on top of the flags in opal_datatype.h */
 /* Is the datatype predefined as MPI type (not necessarily as OPAL type, e.g. struct/block types) */
 #define OMPI_DATATYPE_FLAG_PREDEFINED    0x0200
+#define OMPI_DATATYPE_FLAG_ANALYZED      0x0400
+#define OMPI_DATATYPE_FLAG_MONOTONIC     0x0800
 /* Keep trace of the type of the predefined datatypes */
 #define OMPI_DATATYPE_FLAG_DATA_INT      0x1000
 #define OMPI_DATATYPE_FLAG_DATA_FLOAT    0x2000
@@ -94,7 +97,7 @@ OMPI_DECLSPEC OBJ_CLASS_DECLARATION(ompi_datatype_t);
 /* Using set constant for padding of the DATATYPE handles because the size of
  * base structure is very close to being the same no matter the bitness.
  */
-#define PREDEFINED_DATATYPE_PAD (512)
+#define PREDEFINED_DATATYPE_PAD 512
 
 struct ompi_predefined_datatype_t {
     struct ompi_datatype_t dt;
@@ -151,13 +154,23 @@ ompi_datatype_is_contiguous_memory_layout( const ompi_datatype_t* type, int32_t
     return opal_datatype_is_contiguous_memory_layout(&type->super, count);
 }
 
+static inline int32_t
+ompi_datatype_is_monotonic( ompi_datatype_t * type ) {
+    if (!(type->super.flags & OMPI_DATATYPE_FLAG_ANALYZED)) {
+        if (opal_datatype_is_monotonic(&type->super)) {
+            type->super.flags |= OMPI_DATATYPE_FLAG_MONOTONIC;
+        }
+        type->super.flags |= OMPI_DATATYPE_FLAG_ANALYZED;
+    }
+    return type->super.flags & OMPI_DATATYPE_FLAG_MONOTONIC;
+}
+
 static inline int32_t
 ompi_datatype_commit( ompi_datatype_t ** type )
 {
     return opal_datatype_commit ( (opal_datatype_t*)*type );
 }
 
-
 OMPI_DECLSPEC int32_t ompi_datatype_destroy( ompi_datatype_t** type);
 
 
@@ -165,8 +178,8 @@ OMPI_DECLSPEC int32_t ompi_datatype_destroy( ompi_datatype_t** type);
  * Datatype creation functions
  */
 static inline int32_t
-ompi_datatype_add( ompi_datatype_t* pdtBase, const ompi_datatype_t* pdtAdd, uint32_t count,
-                   OPAL_PTRDIFF_TYPE disp, OPAL_PTRDIFF_TYPE extent )
+ompi_datatype_add( ompi_datatype_t* pdtBase, const ompi_datatype_t* pdtAdd, size_t count,
+                   ptrdiff_t disp, ptrdiff_t extent )
 {
     return opal_datatype_add( &pdtBase->super, &pdtAdd->super, count, disp, extent );
 }
@@ -177,17 +190,17 @@ ompi_datatype_duplicate( const ompi_datatype_t* oldType, ompi_datatype_t** newTy
 OMPI_DECLSPEC int32_t ompi_datatype_create_contiguous( int count, const ompi_datatype_t* oldType, ompi_datatype_t** newType );
 OMPI_DECLSPEC int32_t ompi_datatype_create_vector( int count, int bLength, int stride,
                                                    const ompi_datatype_t* oldType, ompi_datatype_t** newType );
-OMPI_DECLSPEC int32_t ompi_datatype_create_hvector( int count, int bLength, OPAL_PTRDIFF_TYPE stride,
+OMPI_DECLSPEC int32_t ompi_datatype_create_hvector( int count, int bLength, ptrdiff_t stride,
                                                     const ompi_datatype_t* oldType, ompi_datatype_t** newType );
 OMPI_DECLSPEC int32_t ompi_datatype_create_indexed( int count, const int* pBlockLength, const int* pDisp,
                                                     const ompi_datatype_t* oldType, ompi_datatype_t** newType );
-OMPI_DECLSPEC int32_t ompi_datatype_create_hindexed( int count, const int* pBlockLength, const OPAL_PTRDIFF_TYPE* pDisp,
+OMPI_DECLSPEC int32_t ompi_datatype_create_hindexed( int count, const int* pBlockLength, const ptrdiff_t* pDisp,
                                                      const ompi_datatype_t* oldType, ompi_datatype_t** newType );
 OMPI_DECLSPEC int32_t ompi_datatype_create_indexed_block( int count, int bLength, const int* pDisp,
                                                           const ompi_datatype_t* oldType, ompi_datatype_t** newType );
-OMPI_DECLSPEC int32_t ompi_datatype_create_hindexed_block( int count, int bLength, const OPAL_PTRDIFF_TYPE* pDisp,
+OMPI_DECLSPEC int32_t ompi_datatype_create_hindexed_block( int count, int bLength, const ptrdiff_t* pDisp,
                                                            const ompi_datatype_t* oldType, ompi_datatype_t** newType );
-OMPI_DECLSPEC int32_t ompi_datatype_create_struct( int count, const int* pBlockLength, const OPAL_PTRDIFF_TYPE* pDisp,
+OMPI_DECLSPEC int32_t ompi_datatype_create_struct( int count, const int* pBlockLength, const ptrdiff_t* pDisp,
                                                    ompi_datatype_t* const* pTypes, ompi_datatype_t** newType );
 OMPI_DECLSPEC int32_t ompi_datatype_create_darray( int size, int rank, int ndims, int const* gsize_array,
                                                    int const* distrib_array, int const* darg_array,
@@ -198,8 +211,8 @@ OMPI_DECLSPEC int32_t ompi_datatype_create_subarray(int ndims, int const* size_a
                                                     const ompi_datatype_t* oldtype, ompi_datatype_t** newtype);
 static inline int32_t
 ompi_datatype_create_resized( const ompi_datatype_t* oldType,
-                              OPAL_PTRDIFF_TYPE lb,
-                              OPAL_PTRDIFF_TYPE extent,
+                              ptrdiff_t lb,
+                              ptrdiff_t extent,
                               ompi_datatype_t** newType )
 {
     ompi_datatype_t * type;
@@ -213,13 +226,13 @@ ompi_datatype_create_resized( const ompi_datatype_t* oldType,
 }
 
 static inline int32_t
-ompi_datatype_type_lb( const ompi_datatype_t* type, OPAL_PTRDIFF_TYPE* disp )
+ompi_datatype_type_lb( const ompi_datatype_t* type, ptrdiff_t* disp )
 {
     return opal_datatype_type_lb(&type->super, disp);
 }
 
 static inline int32_t
-ompi_datatype_type_ub( const ompi_datatype_t* type, OPAL_PTRDIFF_TYPE* disp )
+ompi_datatype_type_ub( const ompi_datatype_t* type, ptrdiff_t* disp )
 {
     return opal_datatype_type_ub( &type->super, disp);
 }
@@ -231,19 +244,19 @@ ompi_datatype_type_size ( const ompi_datatype_t* type, size_t *size )
 }
 
 static inline int32_t
-ompi_datatype_type_extent( const ompi_datatype_t* type, OPAL_PTRDIFF_TYPE* extent )
+ompi_datatype_type_extent( const ompi_datatype_t* type, ptrdiff_t* extent )
 {
     return opal_datatype_type_extent( &type->super, extent);
 }
 
 static inline int32_t
-ompi_datatype_get_extent( const ompi_datatype_t* type, OPAL_PTRDIFF_TYPE* lb, OPAL_PTRDIFF_TYPE* extent)
+ompi_datatype_get_extent( const ompi_datatype_t* type, ptrdiff_t* lb, ptrdiff_t* extent)
 {
     return opal_datatype_get_extent( &type->super, lb, extent);
 }
 
 static inline int32_t
-ompi_datatype_get_true_extent( const ompi_datatype_t* type, OPAL_PTRDIFF_TYPE* true_lb, OPAL_PTRDIFF_TYPE* true_extent)
+ompi_datatype_get_true_extent( const ompi_datatype_t* type, ptrdiff_t* true_lb, ptrdiff_t* true_extent)
 {
     return opal_datatype_get_true_extent( &type->super, true_lb, true_extent);
 }
@@ -265,7 +278,7 @@ ompi_datatype_copy_content_same_ddt( const ompi_datatype_t* type, size_t count,
                                      char* pDestBuf, char* pSrcBuf )
 {
     int32_t length, rc;
-    OPAL_PTRDIFF_TYPE extent;
+    ptrdiff_t extent;
 
     ompi_datatype_type_extent( type, &extent );
     while( 0 != count ) {
@@ -294,11 +307,11 @@ OMPI_DECLSPEC int32_t ompi_datatype_sndrcv( const void *sbuf, int32_t scount, co
  */
 OMPI_DECLSPEC int32_t ompi_datatype_get_args( const ompi_datatype_t* pData, int32_t which,
                                               int32_t * ci, int32_t * i,
-                                              int32_t * ca, OPAL_PTRDIFF_TYPE* a,
+                                              int32_t * ca, ptrdiff_t* a,
                                               int32_t * cd, ompi_datatype_t** d, int32_t * type);
 OMPI_DECLSPEC int32_t ompi_datatype_set_args( ompi_datatype_t* pData,
                                               int32_t ci, const int32_t ** i,
-                                              int32_t ca, const OPAL_PTRDIFF_TYPE* a,
+                                              int32_t ca, const ptrdiff_t* a,
                                               int32_t cd, ompi_datatype_t* const * d,int32_t type);
 OMPI_DECLSPEC int32_t ompi_datatype_copy_args( const ompi_datatype_t* source_data,
                                                ompi_datatype_t* dest_data );
@@ -363,5 +376,36 @@ OMPI_DECLSPEC int ompi_datatype_safeguard_pointer_debug_breakpoint( const void*
                                                                     int count );
 #endif  /* OPAL_ENABLE_DEBUG */
 
+OMPI_DECLSPEC int ompi_datatype_pack_external( const char datarep[], const void *inbuf, int incount,
+                                               ompi_datatype_t *datatype, void *outbuf,
+                                               MPI_Aint outsize, MPI_Aint *position);
+
+OMPI_DECLSPEC int ompi_datatype_unpack_external( const char datarep[], const void *inbuf, MPI_Aint insize,
+                                                 MPI_Aint *position, void *outbuf, int outcount,
+                                                 ompi_datatype_t *datatype);
+
+OMPI_DECLSPEC int ompi_datatype_pack_external_size( const char datarep[], int incount,
+                                                    ompi_datatype_t *datatype, MPI_Aint *size);
+
+#define OMPI_DATATYPE_RETAIN(ddt)                                       \
+    {                                                                   \
+        if( !ompi_datatype_is_predefined((ddt)) ) {                     \
+            OBJ_RETAIN((ddt));                                          \
+            OPAL_OUTPUT_VERBOSE((0, 100, "Datatype %p [%s] refcount %d in file %s:%d\n",     \
+                                (void*)(ddt), (ddt)->name, (ddt)->super.super.obj_reference_count, \
+                                __FILE__, __LINE__));                   \
+        }                                                               \
+    }
+
+#define OMPI_DATATYPE_RELEASE(ddt)                                      \
+    {                                                                   \
+        if( !ompi_datatype_is_predefined((ddt)) ) {                     \
+            OPAL_OUTPUT_VERBOSE((0, 100, "Datatype %p [%s] refcount %d in file %s:%d\n",     \
+                                (void*)(ddt), (ddt)->name, (ddt)->super.super.obj_reference_count, \
+                                __func__, __LINE__));                   \
+            OBJ_RELEASE((ddt));                                         \
+        }                                                               \
+    }
+
 END_C_DECLS
 #endif  /* OMPI_DATATYPE_H_HAS_BEEN_INCLUDED */
