@@ -1,8 +1,10 @@
 /*
- * Copyright (C) Mellanox Technologies Ltd. 2001-2011.  ALL RIGHTS RESERVED.
+ * Copyright (C) 2001-2011 Mellanox Technologies Ltd. 2001-2011.  ALL RIGHTS RESERVED.
  * Copyright (c) 2016      The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
+ * Copyright (c) 2018      Research Organization for Information Science
+ *                         and Technology (RIST).  All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -14,8 +16,10 @@
 
 #include "opal/runtime/opal.h"
 #include "opal/mca/pmix/pmix.h"
+#include "ompi/attribute/attribute.h"
 #include "ompi/message/message.h"
 #include "ompi/mca/pml/base/pml_base_bsend.h"
+#include "opal/mca/common/ucx/common_ucx.h"
 #include "pml_ucx_request.h"
 
 #include <inttypes.h>
@@ -46,42 +50,78 @@
 #define MODEX_KEY "pml-ucx"
 
 mca_pml_ucx_module_t ompi_pml_ucx = {
-    {
-        mca_pml_ucx_add_procs,
-        mca_pml_ucx_del_procs,
-        mca_pml_ucx_enable,
-        NULL,
-        mca_pml_ucx_add_comm,
-        mca_pml_ucx_del_comm,
-        mca_pml_ucx_irecv_init,
-        mca_pml_ucx_irecv,
-        mca_pml_ucx_recv,
-        mca_pml_ucx_isend_init,
-        mca_pml_ucx_isend,
-        mca_pml_ucx_send,
-        mca_pml_ucx_iprobe,
-        mca_pml_ucx_probe,
-        mca_pml_ucx_start,
-        mca_pml_ucx_improbe,
-        mca_pml_ucx_mprobe,
-        mca_pml_ucx_imrecv,
-        mca_pml_ucx_mrecv,
-        mca_pml_ucx_dump,
-        NULL, /* FT */
-        1ul << (PML_UCX_CONTEXT_BITS),
-        1ul << (PML_UCX_TAG_BITS - 1),
+    .super = {
+        .pml_add_procs     = mca_pml_ucx_add_procs,
+        .pml_del_procs     = mca_pml_ucx_del_procs,
+        .pml_enable        = mca_pml_ucx_enable,
+        .pml_progress      = NULL,
+        .pml_add_comm      = mca_pml_ucx_add_comm,
+        .pml_del_comm      = mca_pml_ucx_del_comm,
+        .pml_irecv_init    = mca_pml_ucx_irecv_init,
+        .pml_irecv         = mca_pml_ucx_irecv,
+        .pml_recv          = mca_pml_ucx_recv,
+        .pml_isend_init    = mca_pml_ucx_isend_init,
+        .pml_isend         = mca_pml_ucx_isend,
+        .pml_send          = mca_pml_ucx_send,
+        .pml_iprobe        = mca_pml_ucx_iprobe,
+        .pml_probe         = mca_pml_ucx_probe,
+        .pml_start         = mca_pml_ucx_start,
+        .pml_improbe       = mca_pml_ucx_improbe,
+        .pml_mprobe        = mca_pml_ucx_mprobe,
+        .pml_imrecv        = mca_pml_ucx_imrecv,
+        .pml_mrecv         = mca_pml_ucx_mrecv,
+        .pml_dump          = mca_pml_ucx_dump,
+        .pml_ft_event      = NULL,
+        .pml_max_contextid = 1ul << (PML_UCX_CONTEXT_BITS),
+        .pml_max_tag       = 1ul << (PML_UCX_TAG_BITS - 1)
     },
-    NULL,   /* ucp_context */
-    NULL    /* ucp_worker */
+    .ucp_context           = NULL,
+    .ucp_worker            = NULL
 };
 
 #define PML_UCX_REQ_ALLOCA() \
     ((char *)alloca(ompi_pml_ucx.request_size) + ompi_pml_ucx.request_size);
 
+#if HAVE_UCP_WORKER_ADDRESS_FLAGS
+static int mca_pml_ucx_send_worker_address_type(int addr_flags, int modex_scope)
+{
+    ucs_status_t status;
+    ucp_worker_attr_t attrs;
+    int rc;
+
+    attrs.field_mask    = UCP_WORKER_ATTR_FIELD_ADDRESS |
+                          UCP_WORKER_ATTR_FIELD_ADDRESS_FLAGS;
+    attrs.address_flags = addr_flags;
+
+    status = ucp_worker_query(ompi_pml_ucx.ucp_worker, &attrs);
+    if (UCS_OK != status) {
+        PML_UCX_ERROR("Failed to query UCP worker address");
+        return OMPI_ERROR;
+    }
+
+    OPAL_MODEX_SEND(rc, modex_scope, &mca_pml_ucx_component.pmlm_version,
+                    (void*)attrs.address, attrs.address_length);
+
+    ucp_worker_release_address(ompi_pml_ucx.ucp_worker, attrs.address);
+
+    if (OMPI_SUCCESS != rc) {
+        return OMPI_ERROR;
+    }
+
+    PML_UCX_VERBOSE(2, "Pack %s worker address, size %ld",
+                    (modex_scope == OPAL_PMIX_LOCAL) ? "local" : "remote",
+                    attrs.address_length);
+
+    return OMPI_SUCCESS;
+}
+#endif
+
 static int mca_pml_ucx_send_worker_address(void)
 {
-    ucp_address_t *address;
     ucs_status_t status;
+
+#if !HAVE_UCP_WORKER_ADDRESS_FLAGS
+    ucp_address_t *address;
     size_t addrlen;
     int rc;
 
@@ -91,16 +131,35 @@ static int mca_pml_ucx_send_worker_address(void)
         return OMPI_ERROR;
     }
 
+    PML_UCX_VERBOSE(2, "Pack worker address, size %ld", addrlen);
+
     OPAL_MODEX_SEND(rc, OPAL_PMIX_GLOBAL,
                     &mca_pml_ucx_component.pmlm_version, (void*)address, addrlen);
-    if (OMPI_SUCCESS != rc) {
-        PML_UCX_ERROR("Open MPI couldn't distribute EP connection details");
-        return OMPI_ERROR;
-    }
 
     ucp_worker_release_address(ompi_pml_ucx.ucp_worker, address);
 
+    if (OMPI_SUCCESS != rc) {
+        goto err;
+    }
+#else
+    /* Pack just network device addresses for remote node peers */
+    status = mca_pml_ucx_send_worker_address_type(UCP_WORKER_ADDRESS_FLAG_NET_ONLY,
+                                                  OPAL_PMIX_REMOTE);
+    if (UCS_OK != status) {
+        goto err;
+    }
+
+    status = mca_pml_ucx_send_worker_address_type(0, OPAL_PMIX_LOCAL);
+    if (UCS_OK != status) {
+        goto err;
+    }
+#endif
+
     return OMPI_SUCCESS;
+
+err:
+    PML_UCX_ERROR("Open MPI couldn't distribute EP connection details");
+    return OMPI_ERROR;
 }
 
 static int mca_pml_ucx_recv_worker_address(ompi_proc_t *proc,
@@ -111,10 +170,14 @@ static int mca_pml_ucx_recv_worker_address(ompi_proc_t *proc,
 
     *address_p = NULL;
     OPAL_MODEX_RECV(ret, &mca_pml_ucx_component.pmlm_version, &proc->super.proc_name,
-                              (void**)address_p, addrlen_p);
+                    (void**)address_p, addrlen_p);
     if (ret < 0) {
-        PML_UCX_ERROR("Failed to receive EP address");
+        PML_UCX_ERROR("Failed to receive UCX worker address: %s (%d)",
+                      opal_strerror(ret), ret);
     }
+
+    PML_UCX_VERBOSE(2, "Got proc %d address, size %ld",
+                    proc->super.proc_name.vpid, *addrlen_p);
     return ret;
 }
 
@@ -185,9 +248,9 @@ int mca_pml_ucx_close(void)
 int mca_pml_ucx_init(void)
 {
     ucp_worker_params_t params;
-    ucs_status_t status;
     ucp_worker_attr_t attr;
-    int rc;
+    ucs_status_t status;
+    int i, rc;
 
     PML_UCX_VERBOSE(1, "mca_pml_ucx_init");
 
@@ -204,30 +267,34 @@ int mca_pml_ucx_init(void)
                                &ompi_pml_ucx.ucp_worker);
     if (UCS_OK != status) {
         PML_UCX_ERROR("Failed to create UCP worker");
-        return OMPI_ERROR;
+        rc = OMPI_ERROR;
+        goto err;
     }
 
     attr.field_mask = UCP_WORKER_ATTR_FIELD_THREAD_MODE;
     status = ucp_worker_query(ompi_pml_ucx.ucp_worker, &attr);
     if (UCS_OK != status) {
-        ucp_worker_destroy(ompi_pml_ucx.ucp_worker);
-        ompi_pml_ucx.ucp_worker = NULL;
         PML_UCX_ERROR("Failed to query UCP worker thread level");
-        return OMPI_ERROR;
+        rc = OMPI_ERROR;
+        goto err_destroy_worker;
     }
 
-    if (ompi_mpi_thread_multiple && attr.thread_mode != UCS_THREAD_MODE_MULTI) {
+    if (ompi_mpi_thread_multiple && (attr.thread_mode != UCS_THREAD_MODE_MULTI)) {
         /* UCX does not support multithreading, disqualify current PML for now */
         /* TODO: we should let OMPI to fallback to THREAD_SINGLE mode */
-        ucp_worker_destroy(ompi_pml_ucx.ucp_worker);
-        ompi_pml_ucx.ucp_worker = NULL;
         PML_UCX_ERROR("UCP worker does not support MPI_THREAD_MULTIPLE");
-        return OMPI_ERROR;
+        rc = OMPI_ERR_NOT_SUPPORTED;
+        goto err_destroy_worker;
     }
 
     rc = mca_pml_ucx_send_worker_address();
     if (rc < 0) {
-        return rc;
+        goto err_destroy_worker;
+    }
+
+    ompi_pml_ucx.datatype_attr_keyval = MPI_KEYVAL_INVALID;
+    for (i = 0; i < OMPI_DATATYPE_MAX_PREDEFINED; ++i) {
+        ompi_pml_ucx.predefined_types[i] = PML_UCX_DATATYPE_INVALID;
     }
 
     /* Initialize the free lists */
@@ -243,15 +310,34 @@ int mca_pml_ucx_init(void)
     PML_UCX_VERBOSE(2, "created ucp context %p, worker %p",
                     (void *)ompi_pml_ucx.ucp_context,
                     (void *)ompi_pml_ucx.ucp_worker);
-    return OMPI_SUCCESS;
+    return rc;
+
+err_destroy_worker:
+    ucp_worker_destroy(ompi_pml_ucx.ucp_worker);
+    ompi_pml_ucx.ucp_worker = NULL;
+err:
+    return OMPI_ERROR;
 }
 
 int mca_pml_ucx_cleanup(void)
 {
+    int i;
+
     PML_UCX_VERBOSE(1, "mca_pml_ucx_cleanup");
 
     opal_progress_unregister(mca_pml_ucx_progress);
 
+    if (ompi_pml_ucx.datatype_attr_keyval != MPI_KEYVAL_INVALID) {
+        ompi_attr_free_keyval(TYPE_ATTR, &ompi_pml_ucx.datatype_attr_keyval, false);
+    }
+
+    for (i = 0; i < OMPI_DATATYPE_MAX_PREDEFINED; ++i) {
+        if (ompi_pml_ucx.predefined_types[i] != PML_UCX_DATATYPE_INVALID) {
+            ucp_dt_destroy(ompi_pml_ucx.predefined_types[i]);
+            ompi_pml_ucx.predefined_types[i] = PML_UCX_DATATYPE_INVALID;
+        }
+    }
+
     ompi_pml_ucx.completed_send_req.req_state = OMPI_REQUEST_INVALID;
     OMPI_REQUEST_FINI(&ompi_pml_ucx.completed_send_req);
     OBJ_DESTRUCT(&ompi_pml_ucx.completed_send_req);
@@ -267,7 +353,7 @@ int mca_pml_ucx_cleanup(void)
     return OMPI_SUCCESS;
 }
 
-ucp_ep_h mca_pml_ucx_add_proc(ompi_communicator_t *comm, int dst)
+static ucp_ep_h mca_pml_ucx_add_proc_common(ompi_proc_t *proc)
 {
     ucp_ep_params_t ep_params;
     ucp_address_t *address;
@@ -276,23 +362,12 @@ ucp_ep_h mca_pml_ucx_add_proc(ompi_communicator_t *comm, int dst)
     ucp_ep_h ep;
     int ret;
 
-    ompi_proc_t *proc0      = ompi_comm_peer_lookup(comm, 0);
-    ompi_proc_t *proc_peer = ompi_comm_peer_lookup(comm, dst);
-
-    /* Note, mca_pml_base_pml_check_selected, doesn't use 3rd argument */
-    if (OMPI_SUCCESS != (ret = mca_pml_base_pml_check_selected("ucx",
-                                                              &proc0,
-                                                              dst))) {
-        return NULL;
-    }
-
-    ret = mca_pml_ucx_recv_worker_address(proc_peer, &address, &addrlen);
+    ret = mca_pml_ucx_recv_worker_address(proc, &address, &addrlen);
     if (ret < 0) {
-        PML_UCX_ERROR("Failed to receive worker address from proc: %d", proc_peer->super.proc_name.vpid);
         return NULL;
     }
 
-    PML_UCX_VERBOSE(2, "connecting to proc. %d", proc_peer->super.proc_name.vpid);
+    PML_UCX_VERBOSE(2, "connecting to proc. %d", proc->super.proc_name.vpid);
 
     ep_params.field_mask = UCP_EP_PARAM_FIELD_REMOTE_ADDRESS;
     ep_params.address    = address;
@@ -300,146 +375,127 @@ ucp_ep_h mca_pml_ucx_add_proc(ompi_communicator_t *comm, int dst)
     status = ucp_ep_create(ompi_pml_ucx.ucp_worker, &ep_params, &ep);
     free(address);
     if (UCS_OK != status) {
-        PML_UCX_ERROR("Failed to connect to proc: %d, %s", proc_peer->super.proc_name.vpid,
-                                                           ucs_status_string(status));
+        PML_UCX_ERROR("ucp_ep_create(proc=%d) failed: %s",
+                      proc->super.proc_name.vpid,
+                      ucs_status_string(status));
         return NULL;
     }
 
-    proc_peer->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML] = ep;
-
+    proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML] = ep;
     return ep;
 }
 
+static ucp_ep_h mca_pml_ucx_add_proc(ompi_communicator_t *comm, int dst)
+{
+    ompi_proc_t *proc0     = ompi_comm_peer_lookup(comm, 0);
+    ompi_proc_t *proc_peer = ompi_comm_peer_lookup(comm, dst);
+    int ret;
+
+    /* Note, mca_pml_base_pml_check_selected, doesn't use 3rd argument */
+    if (OMPI_SUCCESS != (ret = mca_pml_base_pml_check_selected("ucx",
+                                                               &proc0,
+                                                               dst))) {
+        return NULL;
+    }
+
+    return mca_pml_ucx_add_proc_common(proc_peer);
+}
+
 int mca_pml_ucx_add_procs(struct ompi_proc_t **procs, size_t nprocs)
 {
-    ucp_ep_params_t ep_params;
-    ucp_address_t *address;
-    ucs_status_t status;
     ompi_proc_t *proc;
-    size_t addrlen;
     ucp_ep_h ep;
     size_t i;
     int ret;
 
     if (OMPI_SUCCESS != (ret = mca_pml_base_pml_check_selected("ucx",
-                                                              procs,
-                                                              nprocs))) {
+                                                               procs,
+                                                               nprocs))) {
         return ret;
     }
 
     for (i = 0; i < nprocs; ++i) {
         proc = procs[(i + OMPI_PROC_MY_NAME->vpid) % nprocs];
-
-        ret = mca_pml_ucx_recv_worker_address(proc, &address, &addrlen);
-        if (ret < 0) {
-            PML_UCX_ERROR("Failed to receive worker address from proc: %d",
-                          proc->super.proc_name.vpid);
-            return ret;
-        }
-
-        if (proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML]) {
-            PML_UCX_VERBOSE(3, "already connected to proc. %d", proc->super.proc_name.vpid);
-            continue;
-        }
-
-        PML_UCX_VERBOSE(2, "connecting to proc. %d", proc->super.proc_name.vpid);
-
-        ep_params.field_mask = UCP_EP_PARAM_FIELD_REMOTE_ADDRESS;
-        ep_params.address    = address;
-
-        status = ucp_ep_create(ompi_pml_ucx.ucp_worker, &ep_params, &ep);
-        free(address);
-
-        if (UCS_OK != status) {
-            PML_UCX_ERROR("Failed to connect to proc: %d, %s", proc->super.proc_name.vpid,
-                          ucs_status_string(status));
+        ep = mca_pml_ucx_add_proc_common(proc);
+        if (ep == NULL) {
             return OMPI_ERROR;
         }
-
-        proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML] = ep;
     }
 
+    opal_common_ucx_mca_proc_added();
     return OMPI_SUCCESS;
 }
 
-static void mca_pml_ucx_waitall(void **reqs, size_t *count_p)
+static inline ucp_ep_h mca_pml_ucx_get_ep(ompi_communicator_t *comm, int rank)
 {
-    ucs_status_t status;
-    size_t i;
+    ucp_ep_h ep;
 
-    PML_UCX_VERBOSE(2, "waiting for %d disconnect requests", (int)*count_p);
-    for (i = 0; i < *count_p; ++i) {
-        do {
-            opal_progress();
-            status = ucp_request_test(reqs[i], NULL);
-        } while (status == UCS_INPROGRESS);
-        if (status != UCS_OK) {
-            PML_UCX_ERROR("disconnect request failed: %s",
-                          ucs_status_string(status));
-        }
-        ucp_request_free(reqs[i]);
-        reqs[i] = NULL;
+    ep = ompi_comm_peer_lookup(comm, rank)->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML];
+    if (OPAL_LIKELY(ep != NULL)) {
+        return ep;
     }
 
-    *count_p = 0;
+    ep = mca_pml_ucx_add_proc(comm, rank);
+    if (OPAL_LIKELY(ep != NULL)) {
+        return ep;
+    }
+
+    if (rank >= ompi_comm_size(comm)) {
+        PML_UCX_ERROR("Rank number (%d) is larger than communicator size (%d)",
+                      rank, ompi_comm_size(comm));
+    } else {
+        PML_UCX_ERROR("Failed to resolve UCX endpoint for rank %d", rank);
+    }
+
+    return NULL;
 }
 
 int mca_pml_ucx_del_procs(struct ompi_proc_t **procs, size_t nprocs)
 {
     ompi_proc_t *proc;
-    size_t num_reqs, max_reqs;
-    void *dreq, **dreqs;
-    ucp_ep_h ep;
+    opal_common_ucx_del_proc_t *del_procs;
     size_t i;
+    int ret;
 
-    max_reqs = ompi_pml_ucx.num_disconnect;
-    if (max_reqs > nprocs) {
-        max_reqs = nprocs;
-    }
-
-    dreqs = malloc(sizeof(*dreqs) * max_reqs);
-    if (dreqs == NULL) {
+    del_procs = malloc(sizeof(*del_procs) * nprocs);
+    if (del_procs == NULL) {
         return OMPI_ERR_OUT_OF_RESOURCE;
     }
 
-    num_reqs = 0;
-
     for (i = 0; i < nprocs; ++i) {
-        proc = procs[(i + OMPI_PROC_MY_NAME->vpid) % nprocs];
-        ep   = proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML];
-        if (ep == NULL) {
-            continue;
-        }
-
-        PML_UCX_VERBOSE(2, "disconnecting from rank %d", proc->super.proc_name.vpid);
-        dreq = ucp_disconnect_nb(ep);
-        if (dreq != NULL) {
-            if (UCS_PTR_IS_ERR(dreq)) {
-                PML_UCX_ERROR("ucp_disconnect_nb(%d) failed: %s",
-                              proc->super.proc_name.vpid,
-                              ucs_status_string(UCS_PTR_STATUS(dreq)));
-            } else {
-                dreqs[num_reqs++] = dreq;
-            }
-        }
+        proc = procs[i];
+        del_procs[i].ep   = proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML];
+        del_procs[i].vpid = proc->super.proc_name.vpid;
 
+        /* mark peer as disconnected */
         proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML] = NULL;
-
-        if ((int)num_reqs >= ompi_pml_ucx.num_disconnect) {
-            mca_pml_ucx_waitall(dreqs, &num_reqs);
-        }
     }
 
-    mca_pml_ucx_waitall(dreqs, &num_reqs);
-    free(dreqs);
+    ret = opal_common_ucx_del_procs(del_procs, nprocs, OMPI_PROC_MY_NAME->vpid,
+                                    ompi_pml_ucx.num_disconnect, ompi_pml_ucx.ucp_worker);
+    free(del_procs);
 
-    opal_pmix.fence(NULL, 0);
-
-    return OMPI_SUCCESS;
+    return ret;
 }
 
 int mca_pml_ucx_enable(bool enable)
 {
+    ompi_attribute_fn_ptr_union_t copy_fn;
+    ompi_attribute_fn_ptr_union_t del_fn;
+    int ret;
+
+    /* Create a key for adding custom attributes to datatypes */
+    copy_fn.attr_datatype_copy_fn  =
+                    (MPI_Type_internal_copy_attr_function*)MPI_TYPE_NULL_COPY_FN;
+    del_fn.attr_datatype_delete_fn = mca_pml_ucx_datatype_attr_del_fn;
+    ret = ompi_attr_create_keyval(TYPE_ATTR, copy_fn, del_fn,
+                                  &ompi_pml_ucx.datatype_attr_keyval, NULL, 0,
+                                  NULL);
+    if (ret != OMPI_SUCCESS) {
+        PML_UCX_ERROR("Failed to create keyval for UCX datatypes: %d", ret);
+        return ret;
+    }
+
     PML_UCX_FREELIST_INIT(&ompi_pml_ucx.persistent_reqs,
                           mca_pml_ucx_persistent_request_t,
                           128, -1, 128);
@@ -479,11 +535,12 @@ int mca_pml_ucx_irecv_init(void *buf, size_t count, ompi_datatype_t *datatype,
     PML_UCX_TRACE_RECV("irecv_init request *%p=%p", buf, count, datatype, src,
                        tag, comm, (void*)request, (void*)req);
 
-    req->ompi.req_state = OMPI_REQUEST_INACTIVE;
-    req->flags          = 0;
-    req->buffer         = buf;
-    req->count          = count;
-    req->datatype       = mca_pml_ucx_get_datatype(datatype);
+    req->ompi.req_state           = OMPI_REQUEST_INACTIVE;
+    req->ompi.req_mpi_object.comm = comm;
+    req->flags                    = 0;
+    req->buffer                   = buf;
+    req->count                    = count;
+    req->datatype.datatype        = mca_pml_ucx_get_datatype(datatype);
 
     PML_UCX_MAKE_RECV_TAG(req->tag, req->recv.tag_mask, tag, src, comm);
 
@@ -512,7 +569,8 @@ int mca_pml_ucx_irecv(void *buf, size_t count, ompi_datatype_t *datatype,
     }
 
     PML_UCX_VERBOSE(8, "got request %p", (void*)req);
-    *request = req;
+    req->req_mpi_object.comm = comm;
+    *request                 = req;
     return OMPI_SUCCESS;
 }
 
@@ -527,19 +585,19 @@ int mca_pml_ucx_recv(void *buf, size_t count, ompi_datatype_t *datatype, int src
 
     PML_UCX_TRACE_RECV("%s", buf, count, datatype, src, tag, comm, "recv");
 
+    /* coverity[bad_alloc_arithmetic] */
     PML_UCX_MAKE_RECV_TAG(ucp_tag, ucp_tag_mask, tag, src, comm);
     req = PML_UCX_REQ_ALLOCA();
     status = ucp_tag_recv_nbr(ompi_pml_ucx.ucp_worker, buf, count,
                               mca_pml_ucx_get_datatype(datatype),
                               ucp_tag, ucp_tag_mask, req);
 
-    for (;;) {
+    MCA_COMMON_UCX_PROGRESS_LOOP(ompi_pml_ucx.ucp_worker) {
         status = ucp_request_test(req, &info);
         if (status != UCS_INPROGRESS) {
             mca_pml_ucx_set_recv_status_safe(mpi_status, status, &info);
             return OMPI_SUCCESS;
         }
-        opal_progress();
     }
 }
 
@@ -581,22 +639,23 @@ int mca_pml_ucx_isend_init(const void *buf, size_t count, ompi_datatype_t *datat
 
     ep = mca_pml_ucx_get_ep(comm, dst);
     if (OPAL_UNLIKELY(NULL == ep)) {
-        PML_UCX_ERROR("Failed to get ep for rank %d", dst);
         return OMPI_ERROR;
     }
 
-    req->ompi.req_state = OMPI_REQUEST_INACTIVE;
-    req->flags          = MCA_PML_UCX_REQUEST_FLAG_SEND;
-    req->buffer         = (void *)buf;
-    req->count          = count;
-    req->tag            = PML_UCX_MAKE_SEND_TAG(tag, comm);
-    req->send.mode      = mode;
-    req->send.ep        = ep;
+    req->ompi.req_state           = OMPI_REQUEST_INACTIVE;
+    req->ompi.req_mpi_object.comm = comm;
+    req->flags                    = MCA_PML_UCX_REQUEST_FLAG_SEND;
+    req->buffer                   = (void *)buf;
+    req->count                    = count;
+    req->tag                      = PML_UCX_MAKE_SEND_TAG(tag, comm);
+    req->send.mode                = mode;
+    req->send.ep                  = ep;
+
     if (MCA_PML_BASE_SEND_BUFFERED == mode) {
-        req->ompi_datatype = datatype;
+        req->datatype.ompi_datatype = datatype;
         OBJ_RETAIN(datatype);
     } else {
-        req->datatype = mca_pml_ucx_get_datatype(datatype);
+        req->datatype.datatype = mca_pml_ucx_get_datatype(datatype);
     }
 
     *request = &req->ompi;
@@ -695,7 +754,6 @@ int mca_pml_ucx_isend(const void *buf, size_t count, ompi_datatype_t *datatype,
 
     ep = mca_pml_ucx_get_ep(comm, dst);
     if (OPAL_UNLIKELY(NULL == ep)) {
-        PML_UCX_ERROR("Failed to get ep for rank %d", dst);
         return OMPI_ERROR;
     }
 
@@ -710,7 +768,8 @@ int mca_pml_ucx_isend(const void *buf, size_t count, ompi_datatype_t *datatype,
         return OMPI_SUCCESS;
     } else if (!UCS_PTR_IS_ERR(req)) {
         PML_UCX_VERBOSE(8, "got request %p", (void*)req);
-        *request = req;
+        req->req_mpi_object.comm = comm;
+        *request                 = req;
         return OMPI_SUCCESS;
     } else {
         PML_UCX_ERROR("ucx send failed: %s", ucs_status_string(UCS_PTR_STATUS(req)));
@@ -728,16 +787,12 @@ mca_pml_ucx_send_nb(ucp_ep_h ep, const void *buf, size_t count,
 
     req = (ompi_request_t*)mca_pml_ucx_common_send(ep, buf, count, datatype,
                                                    mca_pml_ucx_get_datatype(datatype),
-                                                   tag, mode,
-                                                   mca_pml_ucx_send_completion);
-
+                                                   tag, mode, cb);
     if (OPAL_LIKELY(req == NULL)) {
         return OMPI_SUCCESS;
     } else if (!UCS_PTR_IS_ERR(req)) {
         PML_UCX_VERBOSE(8, "got request %p", (void*)req);
-        ucp_worker_progress(ompi_pml_ucx.ucp_worker);
-        ompi_request_wait(&req, MPI_STATUS_IGNORE);
-        return OMPI_SUCCESS;
+        MCA_COMMON_UCX_WAIT_LOOP(req, ompi_pml_ucx.ucp_worker, "ucx send", ompi_request_free(&req));
     } else {
         PML_UCX_ERROR("ucx send failed: %s", ucs_status_string(UCS_PTR_STATUS(req)));
         return OMPI_ERROR;
@@ -750,21 +805,17 @@ mca_pml_ucx_send_nbr(ucp_ep_h ep, const void *buf, size_t count,
                      ucp_datatype_t ucx_datatype, ucp_tag_t tag)
 
 {
-    void *req;
+    ucs_status_ptr_t req;
     ucs_status_t status;
 
+    /* coverity[bad_alloc_arithmetic] */
     req    = PML_UCX_REQ_ALLOCA();
     status = ucp_tag_send_nbr(ep, buf, count, ucx_datatype, tag, req);
     if (OPAL_LIKELY(status == UCS_OK)) {
         return OMPI_SUCCESS;
     }
 
-    ucp_worker_progress(ompi_pml_ucx.ucp_worker);
-    while ((status = ucp_request_check_status(req)) == UCS_INPROGRESS) {
-        opal_progress();
-    }
-
-    return OPAL_LIKELY(UCS_OK == status) ? OMPI_SUCCESS : OMPI_ERROR;
+    MCA_COMMON_UCX_WAIT_LOOP(req, ompi_pml_ucx.ucp_worker, "ucx send", (void)0);
 }
 #endif
 
@@ -779,7 +830,6 @@ int mca_pml_ucx_send(const void *buf, size_t count, ompi_datatype_t *datatype, i
 
     ep = mca_pml_ucx_get_ep(comm, dst);
     if (OPAL_UNLIKELY(NULL == ep)) {
-        PML_UCX_ERROR("Failed to get ep for rank %d", dst);
         return OMPI_ERROR;
     }
 
@@ -799,8 +849,10 @@ int mca_pml_ucx_send(const void *buf, size_t count, ompi_datatype_t *datatype, i
 }
 
 int mca_pml_ucx_iprobe(int src, int tag, struct ompi_communicator_t* comm,
-                         int *matched, ompi_status_public_t* mpi_status)
+                       int *matched, ompi_status_public_t* mpi_status)
 {
+    static unsigned progress_count = 0;
+
     ucp_tag_t ucp_tag, ucp_tag_mask;
     ucp_tag_recv_info_t info;
     ucp_tag_message_h ucp_msg;
@@ -813,15 +865,16 @@ int mca_pml_ucx_iprobe(int src, int tag, struct ompi_communicator_t* comm,
     if (ucp_msg != NULL) {
         *matched = 1;
         mca_pml_ucx_set_recv_status_safe(mpi_status, UCS_OK, &info);
-    } else {
-        opal_progress();
+    } else  {
+        (++progress_count % opal_common_ucx.progress_iterations) ?
+            (void)ucp_worker_progress(ompi_pml_ucx.ucp_worker) : opal_progress();
         *matched = 0;
     }
     return OMPI_SUCCESS;
 }
 
 int mca_pml_ucx_probe(int src, int tag, struct ompi_communicator_t* comm,
-                        ompi_status_public_t* mpi_status)
+                      ompi_status_public_t* mpi_status)
 {
     ucp_tag_t ucp_tag, ucp_tag_mask;
     ucp_tag_recv_info_t info;
@@ -830,22 +883,23 @@ int mca_pml_ucx_probe(int src, int tag, struct ompi_communicator_t* comm,
     PML_UCX_TRACE_PROBE("probe", src, tag, comm);
 
     PML_UCX_MAKE_RECV_TAG(ucp_tag, ucp_tag_mask, tag, src, comm);
-    for (;;) {
-        ucp_msg = ucp_tag_probe_nb(ompi_pml_ucx.ucp_worker, ucp_tag, ucp_tag_mask,
-                                   0, &info);
+
+    MCA_COMMON_UCX_PROGRESS_LOOP(ompi_pml_ucx.ucp_worker) {
+        ucp_msg = ucp_tag_probe_nb(ompi_pml_ucx.ucp_worker, ucp_tag,
+                                   ucp_tag_mask, 0, &info);
         if (ucp_msg != NULL) {
             mca_pml_ucx_set_recv_status_safe(mpi_status, UCS_OK, &info);
             return OMPI_SUCCESS;
         }
-
-        opal_progress();
     }
 }
 
 int mca_pml_ucx_improbe(int src, int tag, struct ompi_communicator_t* comm,
-                          int *matched, struct ompi_message_t **message,
-                          ompi_status_public_t* mpi_status)
+                        int *matched, struct ompi_message_t **message,
+                        ompi_status_public_t* mpi_status)
 {
+    static unsigned progress_count = 0;
+
     ucp_tag_t ucp_tag, ucp_tag_mask;
     ucp_tag_recv_info_t info;
     ucp_tag_message_h ucp_msg;
@@ -861,7 +915,8 @@ int mca_pml_ucx_improbe(int src, int tag, struct ompi_communicator_t* comm,
         *matched         = 1;
         mca_pml_ucx_set_recv_status_safe(mpi_status, UCS_OK, &info);
     } else  {
-        opal_progress();
+        (++progress_count % opal_common_ucx.progress_iterations) ?
+            (void)ucp_worker_progress(ompi_pml_ucx.ucp_worker) : opal_progress();
         *matched = 0;
     }
     return OMPI_SUCCESS;
@@ -878,7 +933,7 @@ int mca_pml_ucx_mprobe(int src, int tag, struct ompi_communicator_t* comm,
     PML_UCX_TRACE_PROBE("mprobe", src, tag, comm);
 
     PML_UCX_MAKE_RECV_TAG(ucp_tag, ucp_tag_mask, tag, src, comm);
-    for (;;) {
+    MCA_COMMON_UCX_PROGRESS_LOOP(ompi_pml_ucx.ucp_worker) {
         ucp_msg = ucp_tag_probe_nb(ompi_pml_ucx.ucp_worker, ucp_tag, ucp_tag_mask,
                                    1, &info);
         if (ucp_msg != NULL) {
@@ -887,8 +942,6 @@ int mca_pml_ucx_mprobe(int src, int tag, struct ompi_communicator_t* comm,
             mca_pml_ucx_set_recv_status_safe(mpi_status, UCS_OK, &info);
             return OMPI_SUCCESS;
         }
-
-        opal_progress();
     }
 }
 
@@ -960,8 +1013,8 @@ int mca_pml_ucx_start(size_t count, ompi_request_t** requests)
             tmp_req = (ompi_request_t*)mca_pml_ucx_common_send(preq->send.ep,
                                                                preq->buffer,
                                                                preq->count,
-                                                               preq->ompi_datatype,
-                                                               preq->datatype,
+                                                               preq->datatype.ompi_datatype,
+                                                               preq->datatype.datatype,
                                                                preq->tag,
                                                                preq->send.mode,
                                                                mca_pml_ucx_psend_completion);
@@ -969,7 +1022,8 @@ int mca_pml_ucx_start(size_t count, ompi_request_t** requests)
             PML_UCX_VERBOSE(8, "start recv request %p", (void*)preq);
             tmp_req = (ompi_request_t*)ucp_tag_recv_nb(ompi_pml_ucx.ucp_worker,
                                                        preq->buffer, preq->count,
-                                                       preq->datatype, preq->tag,
+                                                       preq->datatype.datatype,
+                                                       preq->tag,
                                                        preq->recv.tag_mask,
                                                        mca_pml_ucx_precv_completion);
         }
