@@ -1,7 +1,7 @@
 /*
  * Copyright (c) 2013      Mellanox Technologies, Inc.
  *                         All rights reserved.
- * Copyright (c) 2014-2016 Research Organization for Information Science
+ * Copyright (c) 2014-2018 Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
  * Copyright (c) 2016      ARM, Inc. All rights reserved.
  * $COPYRIGHT$
@@ -20,6 +20,7 @@
 
 #include "oshmem_config.h"
 #include "opal/datatype/opal_convertor.h"
+#include "opal/mca/common/ucx/common_ucx.h"
 #include "orte/include/orte/types.h"
 #include "orte/runtime/orte_globals.h"
 #include "ompi/datatype/ompi_datatype.h"
@@ -43,43 +44,56 @@
 #define SPML_UCX_PUT_DEBUG    0
 #endif
 
+static
+spml_ucx_mkey_t * mca_spml_ucx_get_mkey_slow(int pe, void *va, void **rva);
+
 mca_spml_ucx_t mca_spml_ucx = {
-    {
+    .super = {
         /* Init mca_spml_base_module_t */
-        mca_spml_ucx_add_procs,
-        mca_spml_ucx_del_procs,
-        mca_spml_ucx_enable,
-        mca_spml_ucx_register,
-        mca_spml_ucx_deregister,
-        mca_spml_base_oob_get_mkeys,
-        mca_spml_ucx_put,
-        mca_spml_ucx_put_nb,
-        mca_spml_ucx_get,
-        mca_spml_ucx_get_nb,
-        mca_spml_ucx_recv,
-        mca_spml_ucx_send,
-        mca_spml_base_wait,
-        mca_spml_base_wait_nb,
-        mca_spml_ucx_quiet, /* At the moment fence is the same as quite for 
-                               every spml */
-        mca_spml_ucx_rmkey_unpack,
-        mca_spml_ucx_rmkey_free,
-        mca_spml_ucx_rmkey_ptr,
-        mca_spml_ucx_memuse_hook,
-        (void*)&mca_spml_ucx
+        .spml_add_procs     = mca_spml_ucx_add_procs,
+        .spml_del_procs     = mca_spml_ucx_del_procs,
+        .spml_enable        = mca_spml_ucx_enable,
+        .spml_register      = mca_spml_ucx_register,
+        .spml_deregister    = mca_spml_ucx_deregister,
+        .spml_oob_get_mkeys = mca_spml_base_oob_get_mkeys,
+        .spml_ctx_create    = mca_spml_ucx_ctx_create,
+        .spml_ctx_destroy   = mca_spml_ucx_ctx_destroy,
+        .spml_put           = mca_spml_ucx_put,
+        .spml_put_nb        = mca_spml_ucx_put_nb,
+        .spml_get           = mca_spml_ucx_get,
+        .spml_get_nb        = mca_spml_ucx_get_nb,
+        .spml_recv          = mca_spml_ucx_recv,
+        .spml_send          = mca_spml_ucx_send,
+        .spml_wait          = mca_spml_base_wait,
+        .spml_wait_nb       = mca_spml_base_wait_nb,
+        .spml_test          = mca_spml_base_test,
+        .spml_fence         = mca_spml_ucx_fence,
+        .spml_quiet         = mca_spml_ucx_quiet,
+        .spml_rmkey_unpack  = mca_spml_ucx_rmkey_unpack,
+        .spml_rmkey_free    = mca_spml_ucx_rmkey_free,
+        .spml_rmkey_ptr     = mca_spml_ucx_rmkey_ptr,
+        .spml_memuse_hook   = mca_spml_ucx_memuse_hook,
+        .self               = (void*)&mca_spml_ucx
     },
 
-    NULL,   /* ucp_context */
-    NULL,   /* ucp_worker */
-    NULL,   /* ucp_peers */
-    0,      /* using_mem_hooks */
-    1,      /* num_disconnect */
-    0       /* heap_reg_nb */
+    .ucp_context            = NULL,
+    .num_disconnect         = 1,
+    .heap_reg_nb            = 0,
+    .enabled                = 0,
+    .get_mkey_slow          = mca_spml_ucx_get_mkey_slow
+};
+
+OBJ_CLASS_INSTANCE(mca_spml_ucx_ctx_list_item_t, opal_list_item_t, NULL, NULL);
+
+mca_spml_ucx_ctx_t mca_spml_ucx_ctx_default = {
+    .ucp_worker = NULL,
+    .ucp_peers  = NULL,
+    .options    = 0
 };
 
 int mca_spml_ucx_enable(bool enable)
 {
-    SPML_VERBOSE(50, "*** ucx ENABLED ****");
+    SPML_UCX_VERBOSE(50, "*** ucx ENABLED ****");
     if (false == enable) {
         return OSHMEM_SUCCESS;
     }
@@ -89,86 +103,44 @@ int mca_spml_ucx_enable(bool enable)
     return OSHMEM_SUCCESS;
 }
 
-
-static void mca_spml_ucx_waitall(void **reqs, size_t *count_p)
-{
-    ucs_status_t status;
-    size_t i;
-
-    SPML_VERBOSE(10, "waiting for %d disconnect requests", *count_p);
-    for (i = 0; i < *count_p; ++i) {
-        do {
-            opal_progress();
-            status = ucp_request_test(reqs[i], NULL);
-        } while (status == UCS_INPROGRESS);
-        if (status != UCS_OK) {
-            SPML_ERROR("disconnect request failed: %s",
-                       ucs_status_string(status));
-        }
-        ucp_request_release(reqs[i]);
-        reqs[i] = NULL;
-    }
-
-    *count_p = 0;
-}
-
 int mca_spml_ucx_del_procs(ompi_proc_t** procs, size_t nprocs)
 {
-    int my_rank = oshmem_my_proc_id();
-    size_t num_reqs, max_reqs;
-    void *dreq, **dreqs;
-    ucp_ep_h ep;
-    size_t i, n;
+    opal_common_ucx_del_proc_t *del_procs;
+    size_t i;
+    int ret;
 
     oshmem_shmem_barrier();
 
-    if (!mca_spml_ucx.ucp_peers) {
+    if (!mca_spml_ucx_ctx_default.ucp_peers) {
         return OSHMEM_SUCCESS;
     }
 
-    max_reqs = mca_spml_ucx.num_disconnect;
-    if (max_reqs > nprocs) {
-        max_reqs = nprocs;
-    }
-
-    dreqs = malloc(sizeof(*dreqs) * max_reqs);
-    if (dreqs == NULL) {
+    del_procs = malloc(sizeof(*del_procs) * nprocs);
+    if (del_procs == NULL) {
         return OMPI_ERR_OUT_OF_RESOURCE;
     }
 
-    num_reqs = 0;
-
     for (i = 0; i < nprocs; ++i) {
-        n  = (i + my_rank) % nprocs;
-        ep = mca_spml_ucx.ucp_peers[n].ucp_conn;
-        if (ep == NULL) {
-            continue;
-        }
+        del_procs[i].ep   = mca_spml_ucx_ctx_default.ucp_peers[i].ucp_conn;
+        del_procs[i].vpid = i;
 
-        SPML_VERBOSE(10, "disconnecting from peer %d", n);
-        dreq = ucp_disconnect_nb(ep);
-        if (dreq != NULL) {
-            if (UCS_PTR_IS_ERR(dreq)) {
-                SPML_ERROR("ucp_disconnect_nb(%d) failed: %s", n,
-                           ucs_status_string(UCS_PTR_STATUS(dreq)));
-            } else {
-                dreqs[num_reqs++] = dreq;
-            }
-        }
-
-        mca_spml_ucx.ucp_peers[n].ucp_conn = NULL;
-
-        if ((int)num_reqs >= mca_spml_ucx.num_disconnect) {
-            mca_spml_ucx_waitall(dreqs, &num_reqs);
-        }
+        /* mark peer as disconnected */
+        mca_spml_ucx_ctx_default.ucp_peers[i].ucp_conn = NULL;
     }
 
-    mca_spml_ucx_waitall(dreqs, &num_reqs);
-    free(dreqs);
+    ret = opal_common_ucx_del_procs(del_procs, nprocs, oshmem_my_proc_id(),
+                                    mca_spml_ucx.num_disconnect,
+                                    mca_spml_ucx_ctx_default.ucp_worker);
 
-    opal_pmix.fence(NULL, 0);
-    free(mca_spml_ucx.ucp_peers);
-    return OSHMEM_SUCCESS;
+    free(del_procs);
+    free(mca_spml_ucx.remote_addrs_tbl);
+    free(mca_spml_ucx_ctx_default.ucp_peers);
+
+    mca_spml_ucx_ctx_default.ucp_peers = NULL;
+
+    opal_common_ucx_mca_proc_added();
+
+    return ret;
 }
 
 /* TODO: move func into common place, use it with rkey exchng too */
@@ -254,30 +226,34 @@ int mca_spml_ucx_add_procs(ompi_proc_t** procs, size_t nprocs)
     ucs_status_t err;
     ucp_address_t *wk_local_addr;
     size_t wk_addr_len;
-    int *wk_roffs, *wk_rsizes;
-    char *wk_raddrs;
+    int *wk_roffs = NULL;
+    int *wk_rsizes = NULL;
+    char *wk_raddrs = NULL;
     ucp_ep_params_t ep_params;
 
 
-    mca_spml_ucx.ucp_peers = (ucp_peer_t *) calloc(nprocs, sizeof(*(mca_spml_ucx.ucp_peers)));
-    if (NULL == mca_spml_ucx.ucp_peers) {
+    mca_spml_ucx_ctx_default.ucp_peers = (ucp_peer_t *) calloc(nprocs, sizeof(*(mca_spml_ucx_ctx_default.ucp_peers)));
+    if (NULL == mca_spml_ucx_ctx_default.ucp_peers) {
         goto error;
     }
 
-    err = ucp_worker_get_address(mca_spml_ucx.ucp_worker, &wk_local_addr, &wk_addr_len);
+    err = ucp_worker_get_address(mca_spml_ucx_ctx_default.ucp_worker, &wk_local_addr, &wk_addr_len);
     if (err != UCS_OK) {
         goto error;
     }
     dump_address(my_rank, (char *)wk_local_addr, wk_addr_len);
 
     rc = oshmem_shmem_xchng(wk_local_addr, wk_addr_len, nprocs,
-            (void **)&wk_raddrs, &wk_roffs, &wk_rsizes);
+                            (void **)&wk_raddrs, &wk_roffs, &wk_rsizes);
     if (rc != OSHMEM_SUCCESS) {
         goto error;
     }
 
     opal_progress_register(spml_ucx_progress);
 
+    mca_spml_ucx.remote_addrs_tbl = (char **)calloc(nprocs, sizeof(char *));
+    memset(mca_spml_ucx.remote_addrs_tbl, 0, nprocs * sizeof(char *));
+
     /* Get the EP connection requests for all the processes from modex */
     for (n = 0; n < nprocs; ++n) {
         i = (my_rank + n) % nprocs;
@@ -286,57 +262,63 @@ int mca_spml_ucx_add_procs(ompi_proc_t** procs, size_t nprocs)
         ep_params.field_mask = UCP_EP_PARAM_FIELD_REMOTE_ADDRESS;
         ep_params.address    = (ucp_address_t *)(wk_raddrs + wk_roffs[i]);
 
-        err = ucp_ep_create(mca_spml_ucx.ucp_worker, 
-                            &ep_params,
-                            &mca_spml_ucx.ucp_peers[i].ucp_conn);
+        err = ucp_ep_create(mca_spml_ucx_ctx_default.ucp_worker, &ep_params,
+                            &mca_spml_ucx_ctx_default.ucp_peers[i].ucp_conn);
         if (UCS_OK != err) {
-            SPML_ERROR("ucp_ep_create failed: %s", ucs_status_string(err));
+            SPML_UCX_ERROR("ucp_ep_create(proc=%zu/%zu) failed: %s", n, nprocs,
+                           ucs_status_string(err));
             goto error2;
         }
+
         OSHMEM_PROC_DATA(procs[i])->num_transports = 1;
         OSHMEM_PROC_DATA(procs[i])->transport_ids = spml_ucx_transport_ids;
+
+        mca_spml_ucx.remote_addrs_tbl[i] = (char *)malloc(wk_rsizes[i]);
+        memcpy(mca_spml_ucx.remote_addrs_tbl[i], (char *)(wk_raddrs + wk_roffs[i]),
+               wk_rsizes[i]);
     }
 
-    ucp_worker_release_address(mca_spml_ucx.ucp_worker, wk_local_addr);
+    ucp_worker_release_address(mca_spml_ucx_ctx_default.ucp_worker, wk_local_addr);
     free(wk_raddrs);
     free(wk_rsizes);
     free(wk_roffs);
 
-    SPML_VERBOSE(50, "*** ADDED PROCS ***");
+    SPML_UCX_VERBOSE(50, "*** ADDED PROCS ***");
     return OSHMEM_SUCCESS;
 
 error2:
     for (i = 0; i < nprocs; ++i) {
-         if (mca_spml_ucx.ucp_peers[i].ucp_conn) {
-             ucp_ep_destroy(mca_spml_ucx.ucp_peers[i].ucp_conn);
+         if (mca_spml_ucx_ctx_default.ucp_peers[i].ucp_conn) {
+             ucp_ep_destroy(mca_spml_ucx_ctx_default.ucp_peers[i].ucp_conn);
+         }
+         if (mca_spml_ucx.remote_addrs_tbl[i]) {
+             free(mca_spml_ucx.remote_addrs_tbl[i]);
          }
     }
-    if (mca_spml_ucx.ucp_peers) 
-        free(mca_spml_ucx.ucp_peers);
-    if (wk_raddrs)
-        free(wk_raddrs);
-    if (wk_rsizes)
-        free(wk_rsizes);
-    if (wk_roffs)
-        free(wk_roffs);
-    if (mca_spml_ucx.ucp_peers)
-        free(mca_spml_ucx.ucp_peers);
+    if (mca_spml_ucx_ctx_default.ucp_peers)
+        free(mca_spml_ucx_ctx_default.ucp_peers);
+    if (mca_spml_ucx.remote_addrs_tbl)
+        free(mca_spml_ucx.remote_addrs_tbl);
+    free(wk_raddrs);
+    free(wk_rsizes);
+    free(wk_roffs);
 error:
     rc = OSHMEM_ERR_OUT_OF_RESOURCE;
-    SPML_ERROR("add procs FAILED rc=%d", rc);
+    SPML_UCX_ERROR("add procs FAILED rc=%d", rc);
     return rc;
 
 }
 
 
+static
 spml_ucx_mkey_t * mca_spml_ucx_get_mkey_slow(int pe, void *va, void **rva)
 {
     sshmem_mkey_t *r_mkey;
 
     r_mkey = mca_memheap_base_get_cached_mkey(pe, va, 0, rva);
     if (OPAL_UNLIKELY(!r_mkey)) {
-        SPML_ERROR("pe=%d: %p is not address of symmetric variable",
-                   pe, va);
+        SPML_UCX_ERROR("pe=%d: %p is not address of symmetric variable",
+                       pe, va);
         oshmem_shmem_abort(-1);
         return NULL;
     }
@@ -371,11 +353,11 @@ void *mca_spml_ucx_rmkey_ptr(const void *dst_addr, sshmem_mkey_t *mkey, int pe)
 #endif
 }
 
-static void mca_spml_ucx_cache_mkey(sshmem_mkey_t *mkey, uint32_t segno, int dst_pe)
+static void mca_spml_ucx_cache_mkey(mca_spml_ucx_ctx_t *ucx_ctx, sshmem_mkey_t *mkey, uint32_t segno, int dst_pe)
 {
     ucp_peer_t *peer;
 
-    peer = &mca_spml_ucx.ucp_peers[dst_pe];
+    peer = &(ucx_ctx->ucp_peers[dst_pe]);
     mkey_segment_init(&peer->mkeys[segno].super, mkey, segno);
 }
 
@@ -384,18 +366,18 @@ void mca_spml_ucx_rmkey_unpack(sshmem_mkey_t *mkey, uint32_t segno, int pe, int
     spml_ucx_mkey_t   *ucx_mkey;
     ucs_status_t err;
     
-    ucx_mkey = &mca_spml_ucx.ucp_peers[pe].mkeys[segno].key;
+    ucx_mkey = &mca_spml_ucx_ctx_default.ucp_peers[pe].mkeys[segno].key;
 
-    err = ucp_ep_rkey_unpack(mca_spml_ucx.ucp_peers[pe].ucp_conn,
+    err = ucp_ep_rkey_unpack(mca_spml_ucx_ctx_default.ucp_peers[pe].ucp_conn,
             mkey->u.data, 
             &ucx_mkey->rkey); 
     if (UCS_OK != err) {
-        SPML_ERROR("failed to unpack rkey: %s", ucs_status_string(err));
+        SPML_UCX_ERROR("failed to unpack rkey: %s", ucs_status_string(err));
         goto error_fatal;
     }
 
     mkey->spml_context = ucx_mkey;
-    mca_spml_ucx_cache_mkey(mkey, segno, pe);
+    mca_spml_ucx_cache_mkey(&mca_spml_ucx_ctx_default, mkey, segno, pe);
     return;
 
 error_fatal:
@@ -415,7 +397,7 @@ void mca_spml_ucx_memuse_hook(void *addr, size_t length)
     }
 
     my_pe    = oshmem_my_proc_id();
-    ucx_mkey = &mca_spml_ucx.ucp_peers[my_pe].mkeys[HEAP_SEG_INDEX].key;
+    ucx_mkey = &mca_spml_ucx_ctx_default.ucp_peers[my_pe].mkeys[HEAP_SEG_INDEX].key;
 
     params.field_mask = UCP_MEM_ADVISE_PARAM_FIELD_ADDRESS |
                         UCP_MEM_ADVISE_PARAM_FIELD_LENGTH |
@@ -427,8 +409,8 @@ void mca_spml_ucx_memuse_hook(void *addr, size_t length)
 
     status = ucp_mem_advise(mca_spml_ucx.ucp_context, ucx_mkey->mem_h, &params);
     if (UCS_OK != status) {
-        SPML_ERROR("ucp_mem_advise failed addr %p len %llu : %s",
-                   addr, (unsigned long long)length, ucs_status_string(status));
+        SPML_UCX_ERROR("ucp_mem_advise failed addr %p len %llu : %s",
+                       addr, (unsigned long long)length, ucs_status_string(status));
     }
 }
 
@@ -456,7 +438,7 @@ sshmem_mkey_t *mca_spml_ucx_register(void* addr,
     segno   = memheap_find_segnum(addr);
     mem_seg = memheap_find_seg(segno);
 
-    ucx_mkey = &mca_spml_ucx.ucp_peers[my_pe].mkeys[segno].key;
+    ucx_mkey = &mca_spml_ucx_ctx_default.ucp_peers[my_pe].mkeys[segno].key;
     mkeys[0].spml_context = ucx_mkey;
 
     /* if possible use mem handle already created by ucx allocator */
@@ -488,24 +470,24 @@ sshmem_mkey_t *mca_spml_ucx_register(void* addr,
         goto error_unmap;
     }
     if (len >= 0xffff) {
-        SPML_ERROR("packed rkey is too long: %llu >= %d",
+        SPML_UCX_ERROR("packed rkey is too long: %llu >= %d",
                 (unsigned long long)len,
                 0xffff);
         oshmem_shmem_abort(-1);
     }
 
-    status = ucp_ep_rkey_unpack(mca_spml_ucx.ucp_peers[oshmem_group_self->my_pe].ucp_conn,
+    status = ucp_ep_rkey_unpack(mca_spml_ucx_ctx_default.ucp_peers[oshmem_group_self->my_pe].ucp_conn,
                                 mkeys[0].u.data,
                                 &ucx_mkey->rkey);
     if (UCS_OK != status) {
-        SPML_ERROR("failed to unpack rkey");
+        SPML_UCX_ERROR("failed to unpack rkey");
         goto error_unmap;
     }
 
     mkeys[0].len     = len;
     mkeys[0].va_base = addr;
     *count = 1;
-    mca_spml_ucx_cache_mkey(&mkeys[0], segno, my_pe);
+    mca_spml_ucx_cache_mkey(&mca_spml_ucx_ctx_default, &mkeys[0], segno, my_pe);
     return mkeys;
 
 error_unmap:
@@ -521,7 +503,7 @@ int mca_spml_ucx_deregister(sshmem_mkey_t *mkeys)
     spml_ucx_mkey_t   *ucx_mkey;
     map_segment_t *mem_seg;
 
-    MCA_SPML_CALL(fence());
+    MCA_SPML_CALL(quiet(oshmem_ctx_default));
     if (!mkeys)
         return OSHMEM_SUCCESS;
 
@@ -530,6 +512,10 @@ int mca_spml_ucx_deregister(sshmem_mkey_t *mkeys)
 
     mem_seg  = memheap_find_va(mkeys[0].va_base);
     ucx_mkey = (spml_ucx_mkey_t*)mkeys[0].spml_context;
+
+    if (OPAL_UNLIKELY(NULL == mem_seg)) {
+        return OSHMEM_ERROR;
+    }
     
     if (MAP_SEGMENT_ALLOC_UCX != mem_seg->type) {
         ucp_mem_unmap(mca_spml_ucx.ucp_context, ucx_mkey->mem_h);
@@ -543,80 +529,203 @@ int mca_spml_ucx_deregister(sshmem_mkey_t *mkeys)
     return OSHMEM_SUCCESS;
 }
 
-int mca_spml_ucx_get(void *src_addr, size_t size, void *dst_addr, int src)
+int mca_spml_ucx_ctx_create(long options, shmem_ctx_t *ctx)
 {
-    void *rva;
-    ucs_status_t status;
-    spml_ucx_mkey_t *ucx_mkey;
+    mca_spml_ucx_ctx_list_item_t *ctx_item;
+    ucp_worker_params_t params;
+    ucp_ep_params_t ep_params;
+    size_t i, nprocs = oshmem_num_procs();
+    ucs_status_t err;
+    int rc = OSHMEM_ERROR;
 
-    ucx_mkey = mca_spml_ucx_get_mkey(src, src_addr, &rva);
-    status = ucp_get(mca_spml_ucx.ucp_peers[src].ucp_conn, dst_addr, size,
-                     (uint64_t)rva, ucx_mkey->rkey);
+    ctx_item = OBJ_NEW(mca_spml_ucx_ctx_list_item_t);
+    ctx_item->ctx.options = options;
 
-    return ucx_status_to_oshmem(status);
+    params.field_mask  = UCP_WORKER_PARAM_FIELD_THREAD_MODE;
+    if (oshmem_mpi_thread_provided == SHMEM_THREAD_SINGLE || options & SHMEM_CTX_PRIVATE || options & SHMEM_CTX_SERIALIZED) {
+        params.thread_mode = UCS_THREAD_MODE_SINGLE;
+    } else {
+        params.thread_mode = UCS_THREAD_MODE_MULTI;
+    }
+
+    err = ucp_worker_create(mca_spml_ucx.ucp_context, &params,
+                            &ctx_item->ctx.ucp_worker);
+    if (UCS_OK != err) {
+        OBJ_RELEASE(ctx_item);
+        return OSHMEM_ERROR;
+    }
+
+    ctx_item->ctx.ucp_peers = (ucp_peer_t *) calloc(nprocs, sizeof(*(ctx_item->ctx.ucp_peers)));
+    if (NULL == ctx_item->ctx.ucp_peers) {
+        goto error;
+    }
+
+    for (i = 0; i < nprocs; i++) {
+        ep_params.field_mask = UCP_EP_PARAM_FIELD_REMOTE_ADDRESS;
+        ep_params.address    = (ucp_address_t *)(mca_spml_ucx.remote_addrs_tbl[i]);
+        err = ucp_ep_create(ctx_item->ctx.ucp_worker, &ep_params,
+                            &ctx_item->ctx.ucp_peers[i].ucp_conn);
+        if (UCS_OK != err) {
+            SPML_ERROR("ucp_ep_create(proc=%d/%d) failed: %s", i, nprocs,
+                       ucs_status_string(err));
+            goto error2;
+        }
+    }
+
+    SHMEM_MUTEX_LOCK(mca_spml_ucx.internal_mutex);
+
+    opal_list_append(&(mca_spml_ucx.ctx_list), &ctx_item->super);
+
+    SHMEM_MUTEX_UNLOCK(mca_spml_ucx.internal_mutex);
+
+    (*ctx) = (shmem_ctx_t)(&ctx_item->ctx);
+
+    return OSHMEM_SUCCESS;
+
+ error2:
+    for (i = 0; i < nprocs; i++) {
+        if (ctx_item->ctx.ucp_peers[i].ucp_conn) {
+            ucp_ep_destroy(ctx_item->ctx.ucp_peers[i].ucp_conn);
+        }
+    }
+
+    if (ctx_item->ctx.ucp_peers)
+        free(ctx_item->ctx.ucp_peers);
+
+ error:
+    ucp_worker_destroy(ctx_item->ctx.ucp_worker);
+    OBJ_RELEASE(ctx_item);
+    rc = OSHMEM_ERR_OUT_OF_RESOURCE;
+    SPML_ERROR("ctx create FAILED rc=%d", rc);
+    return rc;
 }
 
-int mca_spml_ucx_get_nb(void *src_addr, size_t size, void *dst_addr, int src, void **handle)
+void mca_spml_ucx_ctx_destroy(shmem_ctx_t ctx)
+{
+    mca_spml_ucx_ctx_list_item_t *ctx_item, *next;
+    size_t i, nprocs = oshmem_num_procs();
+
+    MCA_SPML_CALL(quiet(ctx));
+
+    SHMEM_MUTEX_LOCK(mca_spml_ucx.internal_mutex);
+
+    /* delete context object from list */
+    OPAL_LIST_FOREACH_SAFE(ctx_item, next, &(mca_spml_ucx.ctx_list),
+                           mca_spml_ucx_ctx_list_item_t) {
+        if ((shmem_ctx_t)(&ctx_item->ctx) == ctx) {
+            opal_list_remove_item(&(mca_spml_ucx.ctx_list), &ctx_item->super);
+            for (i = 0; i < nprocs; i++) {
+                ucp_ep_destroy(ctx_item->ctx.ucp_peers[i].ucp_conn);
+            }
+            free(ctx_item->ctx.ucp_peers);
+            ucp_worker_destroy(ctx_item->ctx.ucp_worker);
+            OBJ_RELEASE(ctx_item);
+            break;
+        }
+    }
+
+    SHMEM_MUTEX_UNLOCK(mca_spml_ucx.internal_mutex);
+}
+
+int mca_spml_ucx_get(shmem_ctx_t ctx, void *src_addr, size_t size, void *dst_addr, int src)
+{
+    void *rva;
+    spml_ucx_mkey_t *ucx_mkey;
+    mca_spml_ucx_ctx_t *ucx_ctx = (mca_spml_ucx_ctx_t *)ctx;
+#if HAVE_DECL_UCP_GET_NB
+    ucs_status_ptr_t request;
+#else
+    ucs_status_t status;
+#endif
+
+    ucx_mkey = mca_spml_ucx_get_mkey(ucx_ctx, src, src_addr, &rva, &mca_spml_ucx);
+#if HAVE_DECL_UCP_GET_NB
+    request = ucp_get_nb(ucx_ctx->ucp_peers[src].ucp_conn, dst_addr, size,
+                         (uint64_t)rva, ucx_mkey->rkey, opal_common_ucx_empty_complete_cb);
+    return opal_common_ucx_wait_request(request, ucx_ctx->ucp_worker, "ucp_get_nb");
+#else
+    status = ucp_get(ucx_ctx->ucp_peers[src].ucp_conn, dst_addr, size,
+                     (uint64_t)rva, ucx_mkey->rkey);
+    return ucx_status_to_oshmem(status);
+#endif
+}
+
+int mca_spml_ucx_get_nb(shmem_ctx_t ctx, void *src_addr, size_t size, void *dst_addr, int src, void **handle)
 {
     void *rva;
     ucs_status_t status;
     spml_ucx_mkey_t *ucx_mkey;
+    mca_spml_ucx_ctx_t *ucx_ctx = (mca_spml_ucx_ctx_t *)ctx;
 
-    ucx_mkey = mca_spml_ucx_get_mkey(src, src_addr, &rva);
-    status = ucp_get_nbi(mca_spml_ucx.ucp_peers[src].ucp_conn, dst_addr, size,
+    ucx_mkey = mca_spml_ucx_get_mkey(ucx_ctx, src, src_addr, &rva, &mca_spml_ucx);
+    status = ucp_get_nbi(ucx_ctx->ucp_peers[src].ucp_conn, dst_addr, size,
                      (uint64_t)rva, ucx_mkey->rkey);
 
     return ucx_status_to_oshmem_nb(status);
 }
 
-int mca_spml_ucx_put(void* dst_addr, size_t size, void* src_addr, int dst)
+int mca_spml_ucx_put(shmem_ctx_t ctx, void* dst_addr, size_t size, void* src_addr, int dst)
 {
     void *rva;
-    ucs_status_t status;
     spml_ucx_mkey_t *ucx_mkey;
+    mca_spml_ucx_ctx_t *ucx_ctx = (mca_spml_ucx_ctx_t *)ctx;
+#if HAVE_DECL_UCP_PUT_NB
+    ucs_status_ptr_t request;
+#else
+    ucs_status_t status;
+#endif
 
-    ucx_mkey = mca_spml_ucx_get_mkey(dst, dst_addr, &rva);
-    status = ucp_put(mca_spml_ucx.ucp_peers[dst].ucp_conn, src_addr, size,
+    ucx_mkey = mca_spml_ucx_get_mkey(ucx_ctx, dst, dst_addr, &rva, &mca_spml_ucx);
+#if HAVE_DECL_UCP_PUT_NB
+    request = ucp_put_nb(ucx_ctx->ucp_peers[dst].ucp_conn, src_addr, size,
+                         (uint64_t)rva, ucx_mkey->rkey, opal_common_ucx_empty_complete_cb);
+    return opal_common_ucx_wait_request(request, ucx_ctx->ucp_worker, "ucp_put_nb");
+#else
+    status = ucp_put(ucx_ctx->ucp_peers[dst].ucp_conn, src_addr, size,
                      (uint64_t)rva, ucx_mkey->rkey);
-
     return ucx_status_to_oshmem(status);
+#endif
 }
 
-int mca_spml_ucx_put_nb(void* dst_addr, size_t size, void* src_addr, int dst, void **handle)
+int mca_spml_ucx_put_nb(shmem_ctx_t ctx, void* dst_addr, size_t size, void* src_addr, int dst, void **handle)
 {
     void *rva;
     ucs_status_t status;
     spml_ucx_mkey_t *ucx_mkey;
+    mca_spml_ucx_ctx_t *ucx_ctx = (mca_spml_ucx_ctx_t *)ctx;
 
-    ucx_mkey = mca_spml_ucx_get_mkey(dst, dst_addr, &rva);
-    status = ucp_put_nbi(mca_spml_ucx.ucp_peers[dst].ucp_conn, src_addr, size,
+    ucx_mkey = mca_spml_ucx_get_mkey(ucx_ctx, dst, dst_addr, &rva, &mca_spml_ucx);
+    status = ucp_put_nbi(ucx_ctx->ucp_peers[dst].ucp_conn, src_addr, size,
                      (uint64_t)rva, ucx_mkey->rkey);
 
     return ucx_status_to_oshmem_nb(status);
 }
 
-int mca_spml_ucx_fence(void)
+
+
+int mca_spml_ucx_fence(shmem_ctx_t ctx)
 {
     ucs_status_t err;
+    mca_spml_ucx_ctx_t *ucx_ctx = (mca_spml_ucx_ctx_t *)ctx;
 
-    err = ucp_worker_flush(mca_spml_ucx.ucp_worker);
+    err = ucp_worker_fence(ucx_ctx->ucp_worker);
     if (UCS_OK != err) {
-         SPML_ERROR("fence failed: %s", ucs_status_string(err));
+         SPML_UCX_ERROR("fence failed: %s", ucs_status_string(err));
          oshmem_shmem_abort(-1);
          return OSHMEM_ERROR;
     }
     return OSHMEM_SUCCESS;
 }
 
-int mca_spml_ucx_quiet(void)
+int mca_spml_ucx_quiet(shmem_ctx_t ctx)
 {
-    ucs_status_t err;
+    int ret;
+    mca_spml_ucx_ctx_t *ucx_ctx = (mca_spml_ucx_ctx_t *)ctx;
 
-    err = ucp_worker_flush(mca_spml_ucx.ucp_worker);
-    if (UCS_OK != err) {
-         SPML_ERROR("quiet failed: %s", ucs_status_string(err));
+    ret = opal_common_ucx_worker_flush(ucx_ctx->ucp_worker);
+    if (OMPI_SUCCESS != ret) {
          oshmem_shmem_abort(-1);
-         return OSHMEM_ERROR;
+         return ret;
     }
     return OSHMEM_SUCCESS;
 }
